/*
 *  RWPdemo.java
 *
 *  (C) 2004 Esa Hyyti
 *
 */
package rwp;

import java.awt.*;
import java.applet.*;
import java.lang.Math;
import java.util.Vector;

import elib.Loc;
import elib.DoubleBuffer;
import stat.Stat;
import stat.doubleStat;
import stat.Bernoulli;

/*****************************************************
 *
 *  The actual simulator
 *
 *****************************************************/

abstract class Node
{
    Node( Waypoint wp )
    {
        WP = wp;
        P  = new Loc( );
        P0 = new Loc( );
        P1 = new Loc( );
        t0 = 0;
        t1 = 0;
        WP.init( P );
        P0.set( P );
        P1.set( P );
    }

    public double x()   { return P.x; }
    public double y()   { return P.y; }

    public abstract void update_waypoints( Loc P0, Loc P1 );
    
        
    public int advance_time( double t, doubleStat legs )
    {
        int count=0;
        
        if ( t > t1 ) // need new waypoint(s)
        {
            while ( t > t1 && count < 99 )
            {
                double leg;
                
                count++;
                t0   = t1;

                update_waypoints( P0, P1 );

                Loc DP = new Loc( P1 );
                DP.sub( P0 );
                leg = DP.len();
                t1  = t0 + leg; // / velocity->draw( rng );
                if ( legs != null )
                    legs.add( leg );
                //cout << "New waypoint: (" << P1.x << "," << P1.y << ")\n";
            }
            if ( count >= 99 )
                System.out.println( "advance_time: max no. of legs exceed" );

            q = 1.0/(t1-t0);
        }
        
        // (t1-t)*P0 + (t-t0)*P1;
        {
            Loc A = new Loc( P0 );  
            Loc B = new Loc( P1 );  
            A.mul( (t1-t)*q );  
            B.mul( (t-t0)*q );  
            P.set( A );
            P.add( B );
        }
        
        return count;
    }

    public  Waypoint WP;
    public  double   t0, t1;  // departure and arrival times
    public  double   q;
    public  Loc      P0, P1, P;
}


class RWPnode extends Node
{
    RWPnode( Waypoint wp )
    {
        super( wp );
    }
    
    public void update_waypoints( Loc P0, Loc P1 )
    {
        P0.set( P1 );
        WP.draw( P1 );
    }
}


class ILPnode extends Node
{
    ILPnode( Waypoint wp )
    {
        super( wp );
    }
    
    public void update_waypoints( Loc P0, Loc P1 )
    {
        WP.draw( P0 );
        WP.draw( P1 );
    }
}

/****************************************************/

class Neighbours
{
    private RWPmodel  model;
    private double    range;       // transmission range
    private double    range2;      // square of above
    private double    dist2[][];   // node distances
    private int       num;         // no. of nodes
    private boolean   conn_status; // connnectivity status
    
    private boolean   reach1[];      // for determining connectivity
    private boolean   reach2[];
    private boolean   reach3[];
    
    Neighbours( RWPmodel s, double r ) { init( s, r );    }
    Neighbours( RWPmodel s           ) { init( s, 0.01 ); }

    private final void init( RWPmodel s, double r )
    {
        model       =  s;
        num         = -1;
        conn_status = false;
        set_range( r );
    }

    public void set_range( double r )
    {
        range  = r;
        range2 = r*r;
    }
    
    public double range( )
    { 
        return range;
    }
    
    private void allocate_memory( int n )
    {
        dist2  = new double[ n ][ n ];
        reach1 = new boolean[ n ];
        reach2 = new boolean[ n ];
        reach3 = new boolean[ n ];
        num    = n;
    }

    public boolean neighbour( int i, int j )
    {
        if ( num < 2 )
            return true;
        return dist2[i][j] <= range2 ? true : false;
    }

    public boolean connected( )
    {
        return conn_status;
    }
    
    private boolean check_connectivity( )
    {
        int total, found;
        
        if ( num < 2 )
            return true;
        
        reach1[0] = reach2[0] = true;
        for( int i=1; i<num; i++ )
            reach1[i] = reach2[i] = false;
        total = found = 1;
        
        while( found > 0 )
        {
            found = 0;

            // determine the new border
            for( int i=0; i<num; i++ )  // reach3 contains the new border
                reach3[i] = false;
            for( int i=0; i<num; i++ )
            {
                if ( reach2[i] )  // reach2 contains the current border
                {
                    for( int j=0; j<num; j++ )
                    {
                        if ( !reach1[j] && neighbour( i, j ) )
                        {
                            reach1[j] = reach3[j] = true;
                            found++;
                        }
                    }
                }
            }

            total += found;
            
            // swap reach2 and reach3
            boolean tmp[] = reach2;
            reach2 = reach3;
            reach3 = tmp;
        }

        return total == num ? true : false;
    }
    
    public void update( )
    {
        if ( num != model.number_of_nodes( ) )
            allocate_memory( model.number_of_nodes( ) );

        for( int i=0; i<num; i++ )
        {
            Node n1 = model.get_node( i );

            dist2[i][i] = 0;  // for purity

            for( int j=i+1; j<num; j++ )
            {
                Node   n2 = model.get_node( j );
                double d2 = n1.P.dist2( n2.P );
                
                dist2[i][j] = d2;
                dist2[j][i] = d2;
            }
        }

        conn_status = check_connectivity( );
    }    
}

/****************************************************/

class ScreenMap
{
    Waypoint WP;     // for max. dimensions
    int      w,  h;
    int      cx, cy;
    double   q;
    
    ScreenMap( Rectangle rb, Waypoint wp )
    {
        WP   = wp;
        double qx, qy;
        
        w    = rb.width;
        h    = rb.height;
        
        qx   = 0.8 * w / wp.w();
        qy   = 0.8 * h / wp.h();
        q    = qx < qy ? qx : qy;
        
        cx   = (int)Math.round( 1.0*w/2 - q*wp.cx() );
        cy   = (int)Math.round( 1.0*h/2 - q*wp.cy() );
    }

    public int xp( double x ) { return (int)( cx + q*x ); }
    public int yp( double y ) { return (int)( h - cy - q*y ); }
    public int rp( double r ) { return (int)( q*r );      }

    private void drawCircle2( Graphics g, double x, double y, double r, boolean fill )
    {
        int x0 = xp( x-r );
        int x1 = xp( x+r );
        int y0 = yp( y+r );
        int y1 = yp( y-r );
        if ( fill )
            g.fillOval( x0, y0, x1-x0, y1-y0 );
        else
            g.drawOval( x0, y0, x1-x0, y1-y0 );
    }

    // r in screen scale
    private void drawCircleS2( Graphics g, double x, double y, int r, boolean fill )
    {
        int x0 = xp( x ) - r;
        int x1 = xp( x ) + r;
        int y0 = yp( y ) - r;
        int y1 = yp( y ) + r;
        
        if ( fill )
            g.fillOval( x0, y0, x1-x0, y1-y0 );
        else
            g.drawOval( x0, y0, x1-x0, y1-y0 );
    }
    
    public void drawCircle( Graphics g, double x, double y, double r ) 
    {
        drawCircle2( g, x, y, r, false ); 
    }
    
    public void drawDisk( Graphics g, double x, double y, double r ) // r in screen scale!
    {
        drawCircle2( g, x, y, r, true ); 
    }

    public void drawDiskS( Graphics g, double x, double y, int r ) // r in screen scale!
    {
        drawCircleS2( g, x, y, r, true ); 
    }
    
    public void drawLine( Graphics g, double x0, double y0, double x1, double y1 )
    {
        g.drawLine( xp( x0 ), yp( y0 ), xp( x1 ), yp( y1 ) );
    }
}

/****************************************************/

abstract class Waypoint
{
    public abstract double cx();        // center of the moving area
    public abstract double cy();
    public abstract double w();         // width and height of
    public abstract double h();         // the moving area
    public abstract void draw( Loc P );
    public          void init( Loc P )  // initial location
    {
        draw( P );
    }
    
    public abstract void drawBorder( Graphics g, ScreenMap sm );
}

class RWP_on_Disk extends Waypoint
{
    public double cx() { return  0.0; }
    public double cy() { return  0.0; }
    public double w()  { return  2.0; }
    public double h()  { return  2.0; }
        
    public void draw( Loc P )
    {
        P.x = 2;
        
        while ( P.len2() >= 1.0 )
        {
            P.x = 2*Math.random() - 1.0;
            P.y = 2*Math.random() - 1.0;
        }
    }

    public void drawBorder( Graphics g, ScreenMap sm )
    {
        sm.drawCircle( g, 0.0, 0.0, 1.0 );        
    }    
}

class RWPB_on_Disk extends RWP_on_Disk
{
    public void draw( Loc P )
    {
        double theta = 2*Math.PI*Math.random();

        P.x = Math.cos(theta);
        P.y = Math.sin(theta);
    }
}

class RWP_on_Square extends Waypoint
{
    public double cx() { return  0.5; }
    public double cy() { return  0.5; }
    public double w()  { return  1.0; }
    public double h()  { return  1.0; }

    public void draw( Loc P )
    {
        P.x = Math.random();
        P.y = Math.random();
    }

    public void drawBorder( Graphics g, ScreenMap sm )
    {
        sm.drawLine( g, 0.0, 0.0, 1.0, 0.0 );
        sm.drawLine( g, 1.0, 0.0, 1.0, 1.0 );
        sm.drawLine( g, 1.0, 1.0, 0.0, 1.0 );
        sm.drawLine( g, 0.0, 1.0, 0.0, 0.0 );
    }
}

class RWPB_on_Square extends RWP_on_Square
{
    public void draw( Loc P )
    {
        double d = 4.0 * Math.random();

        if      ( d < 1.0 ) { P.x = d;      P.y = 0;     }
        else if ( d < 2.0 ) { P.x = 1.0;    P.y = d-1.0; }
        else if ( d < 3.0 ) { P.x = 3.0-d;  P.y = 1.0;   }
        else                { P.x = 0.0;    P.y = 4.0-d; }
    }    
}


class RWP_on_Triangle extends Waypoint
{
    static final double sqr3  = Math.sqrt( 3 );
    static final double sqr32 = Math.sqrt( 3 ) / 2;
    static final double sqr34 = Math.sqrt( 3 ) / 4;

    public double cx() { return  0.5;   }
    public double cy() { return  sqr34; }
    public double w()  { return  1.0;   }
    public double h()  { return  sqr32; }

    public void draw( Loc P )
    {
        do
        {
            P.x =         Math.random();
            P.y = sqr32 * Math.random();
        }
        while ( ( P.x <= 0.5  && sqr3 * P.x       < P.y ) ||
                ( P.x >= 0.5  && sqr3 * (1.0-P.x) < P.y ) );
    }

    public void drawBorder( Graphics g, ScreenMap sm )
    {
        sm.drawLine( g, 0.0, 0.0,   1.0, 0.0   );
        sm.drawLine( g, 1.0, 0.0,   0.5, sqr32 );
        sm.drawLine( g, 0.5, sqr32, 0.0, 0.0   );
    }
}


class Sierpinski extends RWP_on_Triangle
{
    public void init( Loc P )
    {
        draw_corner( P );
    }

    public void draw( Loc P )
    {
        Loc    N   = new Loc( );

        draw_corner( N );
        P.add( N );
        P.mul( 0.5 );
    }

    private void draw_corner( Loc P )
    {
        double rnd = 3.0 * Math.random();
        
        if      ( rnd < 1.0 ) { P.x = 0.0; P.y = 0.0;   }
        else if ( rnd < 2.0 ) { P.x = 1.0; P.y = 0.0;   }
        else                  { P.x = 0.5; P.y = sqr32; }
    }
    
}

/****************************************************/

class RWPmodel
{
    Vector              nodes;
    private double      t;
    public  int         wpmode;
    public  Waypoint    WP;
    
    RWPmodel( int num )
    {
        nodes  = new Vector( );
        WP     = new RWP_on_Disk( );
        wpmode = Config.WPMODE_DEFAULT;
        reset( num );
    }
    
    private Node new_node( )
    {
        if ( wpmode == Config.WPMODE_RWP )
            return new RWPnode( WP );
        else
            return new ILPnode( WP );
    }
    
    public void reset( int n )
    {
        nodes.removeAllElements( );
        n = n < 1 ? 1 : n;
        for( int i=0; i<n; i++ )
            nodes.addElement( new_node( ) );
        t = 0;
    }
    public void reset( )
    {
        reset( number_of_nodes() );
    }

    public int number_of_nodes( )
    {
        return nodes.size();
    }
    
    public Node get_node( int i )
    {
        return (Node)nodes.elementAt(i);
    }
    
    public void advance_time( double dt, doubleStat legs )
    {
        t = t + dt;
        
        for( int i=0; i<nodes.size(); i++ )
        {
            Node n = (Node)nodes.elementAt(i);
            n.advance_time( t, legs );
        }
    }

    public double current_time( )
    {
        return t;
    }
}


class RWPsimu
{
    public  RWPmodel      model;
    public  Neighbours    Ne;
    public  double        dt;

    private double        T;           // time spent in the current connectivity state
    private boolean       prev_C;      // the previous connectivity state

    // Statistics
    public  doubleStat    legs;
    public  doubleStat    Tc;
    public  doubleStat    Td;
    public  Bernoulli     Pc;

    private void init( int num, double dt, double r )
    {
        model = new RWPmodel( num );      
        Ne    = new Neighbours( model, r );
        legs  = new doubleStat( );
        Tc    = new doubleStat( );
        Td    = new doubleStat( );
        Pc    = new Bernoulli( );

        set_dt( dt );
        Ne.update( );
        reset_stat( );
    }

    RWPsimu( )                                   { init(  20,  0.02, 0.5   ); }
    RWPsimu( int num, double dt, double range )  { init( num,    dt, range ); }

    public void reset_stat( )
    {
        Tc.reset( );
        Td.reset( );
        Pc.reset( );
        legs.reset( );
        T = -1;
    }
    
    public void set_wpmode( int m )
    {
        model.wpmode = m;
        model.reset( );
        Ne.update( );
        reset_stat( );
    }
    
    public void set_nodes( int n )
    {
        model.reset( n );
        Ne.update( );
        reset_stat( );
    }
    
    public void set_range( double d )
    {
        Ne.set_range( d );
        Ne.update( );
        reset_stat( );
    }

    public void set_model( int m )
    {
        switch( m )
        {
            case 0:  model.WP = new RWP_on_Disk( );     break;
            case 1:  model.WP = new RWP_on_Square( );   break;
            case 2:  model.WP = new RWP_on_Triangle( ); break;
            case 3:  model.WP = new RWPB_on_Disk( );    break;
            case 4:  model.WP = new RWPB_on_Square( );  break;
            case 5:  model.WP = new Sierpinski( );      break;
            default: return;
        }
        model.reset( );
        reset_stat( );
    }
    
    public void    set_dt( double d )         { dt = d; Pc.reset( );          }
 
    public double  time( )                    { return model.current_time( ); }
    public boolean connected( )               { return Ne.connected( );       }
    public boolean neighbour( int i, int j )  { return Ne.neighbour( i, j );  }
    
    public void advance_time( )
    {
        if ( T >= 0 )
            T += dt;
        
        model.advance_time( dt, legs );
        Ne.update( );        
        
        boolean curr_C = connected( );   // keep eye on connectivity periods

        Pc.add( curr_C );
        
        if ( curr_C != prev_C )
        {
            if ( T > 0 )
            {
                if ( prev_C )   Tc.add( T );
                else            Td.add( T );
            }
            T = 0;
        }
        prev_C = curr_C;
    }
}



/*****************************************************
 *
 *  User interface components to RWP applet
 *
 *****************************************************/

class SpeedControl
{
    static final int    MAX     = 6;
    static final String NAME[]  = { "Very Slow", "Slow", "Normal", "Quick", "Very Quick", "Max" };
    static final int    DELAY[] = { 100, 50, 20, 10, 5, 0 };
    static final int    DEFAULT = 20;

    static final int    is_choice( String arg )
    {
        for( int i=0; i<MAX; i++ )
            if ( arg.equals( NAME[ i ] ) )
                return DELAY[ i ];
        return -1;
    }
}

/*****************************************************/

class Config
{
    Config( TextField st )
    {
        status         = st;
        pause          = false;
        update         = false;
        draw_links     = true;
        draw_nodes     = true;
        draw_legs      = false;
        draw_waypoints = false;
        delay          = SpeedControl.DEFAULT;
        refresh        = REFRESH_DEFAULT;
    }

    public void set_status( String str )
    {
        status.setText( str );
    }

    TextField status;
    boolean   pause;
    boolean   draw_links;
    boolean   draw_legs;
    boolean   draw_nodes;
    boolean   draw_waypoints;
    int       refresh;     // draw every nth realisation
    int       delay;       // ... between updates, in milliseconds
    boolean   update;      // true if, e.g., no. of nodes changes
    
    // Constants for menu etc
    static final Color  MenuBackground   = Color.lightGray;
    
    static final String PROGRAM_NAME     = "Random Waypoint Simulator v.1.1 by Esa Hyytia, 2004-2005";
    
    static final int    MODEL_MAX        = 6;
    static final int    MODEL_DEFAULT    = 0;
    static final String MODEL_NAME[]     = { "Unit Disk", "Unit Square", "Triangle", "Circle",
                                             "Border of Square", "Sierpinski" };
    
    static final int    NODE_NUM_MAX     = 10;
    static final int    NODE_NUM_DEFAULT = 20;
    static final String NODE_NUM_NAME[]  = { "1", "5", "10", "20", "50", "100", "200", "500", "1000", "2000" };
    
    static final int    RANGE_MAX        = 10;
    static final double RANGE_DEFAULT    = 0.5;
    static final String RANGE_NAME[]     = { "0.1", "0.2", "0.3", "0.4", "0.5", "0.6", "0.7", "0.8", "0.9", "1.0" };

    static final int    DT_MAX           = 7;
    static final double DT_DEFAULT       = 0.01;
    static final String DT_NAME[]        = { "0.001", "0.002", "0.005", "0.01", "0.02", "0.05", "0.1" };

    static final int    REFRESH_MAX      = 7;
    static final int    REFRESH_DEFAULT  = 1;
    static final String REFRESH_NAME[]   = { "1", "2", "4", "10", "100", "1000", "10000" };    

    static final int    WPMODE_RWP       = 0;
    static final int    WPMODE_ILP       = 1;
    static final int    WPMODE_MAX       = 2;
    static final int    WPMODE_DEFAULT   = WPMODE_RWP;
    static final String WPMODE_NAME[]    = { "Random Waypoint", "Independent Legs" };
}

/*****************************************************/

public class RWPdemo extends Applet implements Runnable
{
    private Thread       mythread;
    private RWPsimu      simu;
    private Config       cfg;
    private DrawPanel    draw;
    private TextField    status;
    private int          count;
    
    public void init() 
    {
        status = new TextField( Config.PROGRAM_NAME, 50 );
        status.setEditable( false );
        cfg    = new Config( status );
        simu   = new RWPsimu( Config.NODE_NUM_DEFAULT, Config.DT_DEFAULT, Config.RANGE_DEFAULT );
        draw   = new DrawPanel( simu, cfg );
        count  = 0;
        
        setLayout(new BorderLayout());

        add("Center", draw   );
        add("South",  status );
        add("North",  new TopMenu(  simu, cfg ) );
        add("East",   new SideMenu( simu, cfg ) );
        
        mythread = new Thread( this );
        mythread.start( );
    }

    public String getAppletInfo() 
    {
        return "Random Waypoint Applet by Esa Hyytia 2004";
    }

    public void run() 
    {
        while( true )
        {
            try
            {
                int t = cfg.pause ? 500 : cfg.delay;
                Thread.sleep( t );
            }
            catch ( InterruptedException e ) {}

            if ( !cfg.pause )
            {
                simu.advance_time( );
                count++;
            }
            
            if ( cfg.update || count > cfg.refresh )
            {
                //draw.repaint( );
                draw.paint( draw.getGraphics( ) );
                cfg.update = false;
                count      = 0;
            }
        }
    }
    
    public boolean handleEvent(Event e) 
    {
        switch (e.id) 
        {
            case Event.WINDOW_DESTROY:
                System.exit(0);
                return true;
            default:
                return false;
        }
    }
    
    public static void main(String args[]) 
    {
        Frame   f    = new Frame("RWPdemo");
        RWPdemo rwp  = new RWPdemo();

        rwp.init();
        rwp.start();
        
        f.add("Center", rwp );
        f.setSize(500, 300);
        f.show();
    }
}

/*****************************************************/


class DrawPanel extends Panel 
{
    RWPsimu simu;
    Config  cfg;

    DoubleBuffer db;  // double buffering
    
    public DrawPanel( RWPsimu s, Config c ) 
    {
        simu = s;
        cfg  = c;
        db   = new DoubleBuffer( this );
        setBackground( Config.MenuBackground );
    }

    public boolean handleEvent(Event e) 
    {
        switch (e.id) 
        {
            case Event.MOUSE_DOWN:
                if      ( e.modifiers ==  Event.ALT_MASK )  // middle button
                    ;
                else if ( e.modifiers == Event.META_MASK )  // right button
                    ;
                else
                    ;
                return true;

            case Event.WINDOW_DESTROY:
                System.exit(0);
                return true;
        }
        return false;
    }
    
    public void piirra(Graphics g) 
    {
        ScreenMap sm    = new ScreenMap( getBounds(), simu.model.WP );
        Vector    nodes = simu.model.nodes;
        
        g.setColor(Config.MenuBackground);
        g.fillRect(0, 0, sm.w, sm.h );
        g.draw3DRect(0, 0, sm.w, sm.h, false);
        
        if ( cfg.draw_legs )
        {
            g.setColor( Color.yellow );
            for( int i=0; i<nodes.size(); i++ )
            {
                Node n  = (Node)nodes.elementAt(i);
                Loc  P0 = n.P0;
                Loc  P1 = n.P1;
                
                g.drawLine( sm.xp( P0.x ), sm.yp( P0.y ),
                            sm.xp( P1.x ), sm.yp( P1.y ) );
            }
        }

        g.setColor( Color.black );
        simu.model.WP.drawBorder( g, sm );

        if ( cfg.draw_links )
        {
            g.setColor( Color.blue );
            for( int i=0; i<nodes.size(); i++ )
            {
                Node n1 = (Node)nodes.elementAt(i);
                for( int j=i+1; j<nodes.size(); j++ )
                {
                    if ( simu.neighbour( i, j ) )
                    {
                        Node n2 = (Node)nodes.elementAt(j);
                        
                        g.drawLine( sm.xp( n1.x() ), sm.yp( n1.y() ),
                                    sm.xp( n2.x() ), sm.yp( n2.y() ) );
                    }
                }
            }
        }

        if ( cfg.draw_waypoints )
        {
            g.setColor( Color.black );
            for( int i=0; i<nodes.size(); i++ )
            {
                Node n  = (Node)nodes.elementAt(i);
                Loc  P  = n.P1;
                
                sm.drawDiskS( g, P.x, P.y, 4 );
            }
        }
        
        if ( cfg.draw_nodes )
        {
            g.setColor( Color.red );
            for( int i=0; i<nodes.size(); i++ )
            {
                Node n = (Node)nodes.elementAt(i);
                sm.drawDiskS( g, n.x(), n.y(), 6 );
            }
        }
    }
    
    public void paint(Graphics g)
    {
        Graphics g2 = db.get_graphics();
        
        if ( g2!=null )
        {
            piirra( g2 );
            db.refresh( );
        }
    }
    

}

/*****************************************************/

class StatLabel extends Label
{
    private Stat   stat;
    private String desc;
    
    StatLabel( Stat s, String d )
    {
        stat = s;
        desc = d;
        update( );
    }

    public void update( )
    {
        setText( desc + Double.toString( Math.round( 10000 * stat.mean() ) / 10000.0 ) );
    }
}


class StatPanel extends Panel
{
    RWPsimu       simu;
    Label         Tm, Ct;
    StatLabel     Pc, Lg, Tc, Td;
    
    static final String  Tm_desc = "Current Time : ";
    static final String  Ct_desc = "Current State: ";
    static final String  Title   = "Statistics:";
    static final String  Lg_desc = "  E[ leg ] = ";
    static final String  Pc_desc = "  P[conn.] = ";
    static final String  Tc_desc = "  E[ Tc  ] = ";
    static final String  Td_desc = "  E[ Td  ] = ";
    
    public StatPanel( RWPsimu s )
    {
        simu = s;

        setLayout(new GridLayout(0,1) );
        setBackground( Config.MenuBackground );

        add( (Tm = new Label( Tm_desc ) ) );
        add( (Ct = new Label( Ct_desc ) ) );
        add( new Label( Title ) );
        add( (Lg = new StatLabel( simu.legs, Lg_desc ) ) );
        add( (Pc = new StatLabel( simu.Pc,   Pc_desc ) ) );
        add( (Tc = new StatLabel( simu.Tc,   Tc_desc ) ) );
        add( (Td = new StatLabel( simu.Td,   Td_desc ) ) );
        update( );
    }

    public void update( )
    {
        Lg.update( );
        Pc.update( );
        Tc.update( );
        Td.update( );
        Tm.setText( Tm_desc + Double.toString( Math.round( 10000*simu.time() ) / 10000.0 ) );
        Ct.setText( Ct_desc + (simu.connected() ? "connected" : "disconnected" ) );
    }
}


//
// Presents Information about given Node
class StatDialog extends Dialog 
{
    protected StatPanel     panel;
    private   RWPsimu       simu;
    private   Button        reset;
    private   Button        update;
    private   Button        close;
    
    public StatDialog( Frame parent, RWPsimu s )
    {
        super(parent, "RWP Stat", false);
        simu  = s;
        panel = new StatPanel( simu );
        setLayout(new BorderLayout(15, 15));
        add( "Center", panel );

        Panel p = new Panel( );
        p.add( (reset  = new Button( "Reset"  ) ) );
        p.add( (update = new Button( "Update" ) ) );
        p.add( (close  = new Button( "Close"  ) ) );
        add( "South", p );
        this.pack();
    }

    // Pop down the window when the button is clicked.
    public boolean action(Event e, Object arg)
    {
        if (e.target == close) 
        {
            //panel.close( true );
            this.hide();
            this.dispose();
            return true;
        }
        else if (e.target == update) 
        {
            panel.update( );
            panel.repaint( );
            return true;
        }
        else if (e.target == reset) 
        {
            simu.reset_stat( );
        }
        
        return false;
    }
}

/**********************************************************/

class SideMenu extends Panel 
{
    Config    cfg;
    RWPsimu   simu;

    Choice    wpmode;
    Choice    model;
    Choice    nodenum;
    Choice    range;
    Choice    dt;
    Choice    refresh;

    public SideMenu( RWPsimu s, Config c )
    {
        simu = s;        
        cfg  = c;
        
        setLayout(new FlowLayout() );
        setBackground( Config.MenuBackground );
        setBackground(Config.MenuBackground);

        Panel p = new Panel( );
        p.setLayout(new GridLayout(0,1) );

        Choice        choice;
        
        p.add( new Label( "SIMULATION" ) );

        p.add( new Label( "Model:" ) );

        choice  = new Choice();
        wpmode  = choice;
        for( int i=0; i<Config.WPMODE_MAX; i++ )
            choice.addItem( Config.WPMODE_NAME[i] );
        choice.setBackground( Config.MenuBackground );
        choice.select( 0 );
        p.add( choice );

        choice  = new Choice();
        model = choice;
        for( int i=0; i<Config.MODEL_MAX; i++ )
            choice.addItem( Config.MODEL_NAME[i] );
        choice.setBackground( Config.MenuBackground );
        choice.select( 0 );
        p.add( choice );

        p.add( new Label( "No. of nodes:" ) );
        choice  = new Choice();
        nodenum = choice;
        for( int i=0; i<Config.NODE_NUM_MAX; i++ )
            choice.addItem( Config.NODE_NUM_NAME[i] );
        choice.setBackground( Config.MenuBackground );
        choice.select( 3 );
        p.add( choice );
        
        p.add( new Label( "Range:" ) );
        choice  = new Choice();
        range   = choice;
        for( int i=0; i<Config.RANGE_MAX; i++ )
            choice.addItem( Config.RANGE_NAME[i] );
        choice.setBackground(Config.MenuBackground);
        choice.select( 4 );
        p.add( choice );

        p.add( new Label( "Time Increment:" ) );
        choice  = new Choice();
        dt      = choice;
        for( int i=0; i<Config.DT_MAX; i++ )
            choice.addItem( Config.DT_NAME[i] );
        choice.setBackground(Config.MenuBackground);
        choice.select( 3 );
        p.add( choice );        
        add( p );

        p.add( new Label( "" ) );  // "empty slot"
        p.add( new Label( "VISUALISATION" ) );
        p.add( new Label( "Sim. speed:" ) );
        choice  = new Choice();
        for( int i=0; i<SpeedControl.MAX; i++ )
            choice.addItem( SpeedControl.NAME[i] );
        choice.setBackground(Config.MenuBackground);
        choice.select( 2 );
        p.add( choice );

        p.add( new Label( "Refresh every:" ) );
        choice  = new Choice();
        refresh = choice;
        for( int i=0; i<Config.REFRESH_MAX; i++ )
            choice.addItem( Config.REFRESH_NAME[i] );
        choice.setBackground(Config.MenuBackground);
        choice.select( 0 );
        p.add( choice );
    }
    
    public void paint(Graphics g) 
    { 
        Rectangle r = getBounds();

        g.setColor(Config.MenuBackground);
        g.draw3DRect(0, 0, r.width-1, r.height-1, false);
   }
   
    public boolean action(Event e, Object arg) 
    {
        if (e.target instanceof Choice) 
        {
            String str = (String)arg;

            if ( model == e.target )
            {
                simu.set_model( model.getSelectedIndex() );
                cfg.update = true;
                return true;
            }

            if ( wpmode == e.target )
            {
                simu.set_wpmode( wpmode.getSelectedIndex() );
                cfg.update = true;
                return true;
            }

            if ( nodenum == e.target )
            {
                simu.set_nodes( Integer.parseInt( str ) );
                cfg.update = true;
                return true;
            }

            if ( range == e.target )
            {
                simu.set_range( Double.parseDouble( str ) );
                cfg.update = true;
                return true;
            }

            if ( refresh == e.target )
            {
                cfg.refresh = Integer.parseInt( str );
                return true;
            }


            if ( dt == e.target )
            {
                simu.set_dt( Double.parseDouble( str ) );
                return true;
            }

            int delay = SpeedControl.is_choice( str );
            if ( delay >= 0 )
            {
                cfg.delay = delay;
                return true;
            }
	} 
        return true;
    }
}

class TopMenu extends Panel 
{
    Config    cfg;
    RWPsimu   simu;

    public TopMenu( RWPsimu s, Config c )
    {
        simu = s;
        cfg  = c;
        
        setLayout(new FlowLayout());

        //target.setForeground(Color.red);
        Choice        choice;
        

        add( new Label( "Draw: " ) );
        add( new Checkbox( "nodes",      true  ) );
        add( new Checkbox( "links",      true  ) );
        add( new Checkbox( "legs",       false ) );
        add( new Checkbox( "waypoints",  false ) );

        add( new Button( "Pause" ) );
        add( new Button( "Stats" ) );
        add( new Button( "About" ) );
        setBackground(Config.MenuBackground);
    }

    public void paint(Graphics g) 
    {
        Rectangle r = getBounds();

        g.setColor(Config.MenuBackground);
        g.draw3DRect(0, 0, r.width-1, r.height-1, false);
    }
   
    public boolean action(Event e, Object arg) 
    {
        if (e.target instanceof Button) 
        {
            String choice = (String)arg;
	    if ( choice.equals( "Pause"  ) )
            {
                cfg.pause = !cfg.pause;
                if ( cfg.pause )  cfg.set_status( "Paused" );
                else              cfg.set_status( "Running" );
            }
	    else if ( choice.equals( "Stats"  ) )
            {
                popUp( );
            }
	    else if ( choice.equals( "About"  ) )
                cfg.set_status( Config.PROGRAM_NAME );
        }
        else if (e.target instanceof Checkbox) 
        {
            Checkbox cb  = (Checkbox)e.target;
            String   str = cb.getLabel();

            if      ( str.equals( "links" ) ) 
            {
                cfg.draw_links = cb.getState( );
                cfg.update     = true;
            }
            else if ( str.equals( "legs"  ) )
            {
                cfg.draw_legs = cb.getState( );
                cfg.update    = true;
            }
            else if ( str.equals( "waypoints"  ) )
            {
                cfg.draw_waypoints = cb.getState( );
                cfg.update         = true;
            }
            else if ( str.equals( "nodes"  ) )
            {
                cfg.draw_nodes = cb.getState( );
                cfg.update     = true;
            }
        }
        
        return true;
    }

    public synchronized void popUp( ) 
    {
        /*
          if ( stat_popup != null )
            return;
        */
        
        Frame  f = new Frame( );
        f.setBounds( 50, 50, 600, 400 );
        Dialog d = new StatDialog( f, simu );
        d.show(); 
    }
}
